<?php
/**
 * Altis Acclerate Admin UI.
 */

namespace Altis\Accelerate\Admin;

use Altis\Accelerate;
use Altis\Accelerate\Utils;

const MENU_SLUG = 'altis-accelerate-settings';
const FEEDBACK_FORM_URL = 'https://altis-accelerate.canny.io/';

/**
 * Setup admin hooks.
 *
 * @return void
 */
function bootstrap() : void {
	add_filter( 'plugin_action_links_' . plugin_basename( Accelerate\PLUGIN_FILE ), __NAMESPACE__ . '\\add_plugin_row_links' );

	add_action( 'admin_menu', __NAMESPACE__ . '\\register_admin_menu' );
	add_action( 'admin_menu', __NAMESPACE__ . '\\register_settings_admin_page', 11 );

	register_setting( 'altis-accelerate', 'altis_config', [
		'type' => 'string',
		'show_in_rest' => false,
		'sanitize_callback' => __NAMESPACE__ . '\\sanitize_config_setting',
	] );

	add_filter( 'plugin_row_meta', __NAMESPACE__ . '\\plugin_row_meta', 10, 3 );
	add_filter( 'admin_footer_text', __NAMESPACE__ . '\\render_feedback_link' );

	add_action( 'admin_head', __NAMESPACE__ . '\\action_admin_head' );
	add_action( 'all_admin_notices', __NAMESPACE__ . '\\add_plugin_header' );
	add_action( 'admin_footer', __NAMESPACE__ . '\\add_plugin_footer', 0 );
	add_action( 'admin_enqueue_scripts', __NAMESPACE__ . '\\action_admin_enqueue_scripts' );

	Notices\setup();

	$config = Accelerate\get_config();
	if ( ! $config ) {
		add_action( 'admin_footer', __NAMESPACE__ . '\\render_config_warning' );
	}
}

/**
 * Register scripts for all admin pages.
 */
function action_admin_enqueue_scripts() : void {
	if ( ! is_altis_screen() ) {
		return;
	}

	Utils\register_assets( 'accelerate-admin', [
		'dependencies' => [
			'wp-api-fetch',
			'wp-components',
			'wp-data',
			'wp-element',
			'wp-i18n',
			'wp-url',
			'wp-html-entities',
		]
	], true );
}

/**
 * Check if current admin screen is an Altis screen.
 *
 * @return boolean
 */
function is_altis_screen() : bool {
	$screen = get_current_screen();

	$altis_screens = [
		'toplevel_page_accelerate',
		'edit-xb',
		'edit-wp_block',
		'edit-broadcast',
		'broadcast',
		'accelerate_page_altis-analytics',
		'accelerate_page_audience',
		'accelerate_page_altis-accelerate-settings',
	];

	if ( Accelerate\get_config() ) {
		$altis_screens[] = 'dashboard';
	}

	return in_array( $screen->id, $altis_screens, true );
}

/**
 * Prints the Altis plugin page header.
 */
function add_plugin_header() : void {
	global $title;

	if ( ! is_altis_screen() ) {
		return;
	}

	printf( '
		<div class="altis-ui">
			<div class="Header">
				<h1 class="Logo">
					<img alt="Altis Accelerate" height="26" src="%s" width="64" />
					<span class="Beta">
						%s
					</span>
					<span class="Title">%s</span>
				</h1>
				<div class="CannyWrap">
					<div class="Canny">
						<img src="%s" height="50" width="50" />
						<strong>We are in beta.</strong> Vote on upcoming features or report bugs. <a href="https://altis-accelerate.canny.io/" target="_blank">Talk to us here</a>
					</div>
				</div>
			</div>',
		plugins_url( 'assets/altis-logo.svg', Accelerate\PLUGIN_FILE ),
		__( 'BETA', 'altis' ),
		$title,
		plugins_url( 'assets/canny.png', Accelerate\PLUGIN_FILE ),
		__( 'BETA', 'altis' ),
		$title
	);

	Notices\get_page_notice();
}

/**
 * Closes off the Altis UI div.
 *
 * @param string $data The data to print.
 */
function add_plugin_footer() : void {
	if ( ! is_altis_screen() ) {
		return;
	}

	echo '</div>';
}

/**
 * Fires as an admin screen or script is being initialized.
 */
function action_admin_head() : void {
	if ( ! is_altis_screen() ) {
		return;
	}

	add_notices_wrapper();
}

/**
 * Render the config warning if not configured.
 *
 * @return void
 */
function render_config_warning() : void {
	if ( is_altis_screen() ) {
		return;
	}

	get_activation_message();
}

/**
 * Output the activation message.
 *
 * @return void
 */
function get_activation_message() : void {
	$message = sprintf(
		__( '<strong>Altis Accelerate is not active.</strong> <a href="%s">Add your API key</a> to get started.', 'altis' ),
		admin_url( 'admin.php?page=' . MENU_SLUG )
	);
	printf(
		'<div id="message" class="error"><p>%s</p></div>',
		wp_kses_post( $message )
	);
}

/**
 * Filters the list of action links displayed for a specific plugin in the Plugins list table.
 *
 * @param string[] $links List of plugin action links.
 * @return string[]
 */
function add_plugin_row_links( array $links ) : array {
	$links['settings'] = sprintf(
		'<a href="%s">%s</a>',
		admin_url( 'admin.php?page=' . MENU_SLUG ),
		__( 'Settings', 'altis' )
	);
	return $links;
}

/**
 * Add Accelerate config page.
 *
 * @return void
 */
function register_settings_admin_page() : void {
	add_submenu_page(
		'accelerate',
		_x( 'Settings', 'settings page title', 'altis' ),
		_x( 'Settings', 'settings menu title', 'altis' ),
		'manage_options',
		MENU_SLUG,
		__NAMESPACE__ . '\\render_settings_page',
		1000
	);

	// Register a new section in the "wporg" page.
	add_settings_section(
		'altis-accelerate-setup',
		_x( 'Setup', 'settings section title', 'altis' ),
		'__return_null',
		'altis-accelerate'
	);

	// Register a new field in the "wporg_section_developers" section, inside the "wporg" page.
	add_settings_field(
		'altis_config',
		'API Configuration',
		__NAMESPACE__ . '\\render_config_settings_field',
		'altis-accelerate',
		'altis-accelerate-setup'
	);
}

/**
 * Accelerate settings section description placeholder.
 *
 * @return void
 */
function render_settings_section() : void {

}

/**
 * Config settings field callback.
 *
 * @param array $args Field settings data.
 * @return void
 */
function render_config_settings_field( array $args ) : void {
	?>
	<textarea name="altis_config" style="min-height: 150px" class="widefat"><?php echo esc_textarea( get_option( 'altis_config' ) ) ?></textarea>
	<p class="description">
		<?php esc_html_e( 'Enter your API configuration as provided by Altis.', 'altis' ); ?>
	</p>
	<?php
}

/**
 * Settings page UI.
 *
 * @return void
 */
function render_settings_page() : void {
	settings_errors( 'altis-accelerate' );
	?>
	<div class="wrap">
		<h1><?php echo esc_html( get_admin_page_title() ); ?></h1>
		<form action="options.php" method="post">
			<?php
			settings_fields( 'altis-accelerate' );
			do_settings_sections( 'altis-accelerate' );
			// Output save settings button.
			submit_button( __( 'Save Changes', 'altis' ) );
			?>
		</form>
	</div>
	<?php
}

/**
 * Settings page validation callback.
 *
 * @param string $config The Altis Accelerate config.
 * @return string
 */
function sanitize_config_setting( string $config ) : string {
	$parsed_config = json_decode( $config );
	if ( json_last_error() ) {
		// phpcs:ignore
		add_settings_error( 'altis_config', 'invalid-json', __( sprintf( 'Unable parse JSON from config file. %s.', json_last_error_msg() ), 'altis' ) );
		return '';
	}

	return $config;
}

/**
 * Filters the array of row meta for each plugin in the Plugins list table.
 *
 * @param string[] $plugin_meta An array of the plugin's metadata, including the version, author, author URI, and plugin URI.
 * @param string $plugin_file Path to the plugin file relative to the plugins directory.
 * @param array $plugin_data An array of plugin data.
 * @return string[] An array of the plugin's metadata, including the version, author, author URI, and plugin URI.
 */
function plugin_row_meta( array $plugin_meta, string $plugin_file, array $plugin_data ) : array {
	if ( $plugin_data['UpdateURI'] !== 'https://wordpress.org/plugins/altis-accelerate/' ) {
		return $plugin_meta;
	}

	$plugin_meta[] = sprintf(
		'<a href="%s">%s</a>',
		FEEDBACK_FORM_URL,
		__( 'Send Feedback', 'altis' )
	);

	return $plugin_meta;
}

/**
 * Show feedback form URL in footer.
 *
 * @param string $text The admin footer text to display.
 * @return string
 */
function render_feedback_link( string $text ) : string {
	if ( ! is_altis_screen() ) {
		return $text;
	}

	$message = __( 'Altis Accelerate Feature Requests & Bugs', 'altis' );
	$message = sprintf(
		'<a href="%s" target="altis-accelerate-feedback">%s<span style="font-size:inherit;text-decoration:none;" class="dashicons dashicons-external"></span></a>',
		esc_url( FEEDBACK_FORM_URL ),
		wp_kses_post( $message )
	);

	if ( ! empty ( $text ) ) {
		$message = ' | ' . $message;
	}

	return $text . $message;
}


/**
 * Add a wrapper for admin notices on custom admin pages.
 *
 * @return void
 */
function add_notices_wrapper() : void {
	// High priority so the opening div is before any notices.
	add_action( 'admin_notices', __NAMESPACE__ . '\\add_notices_wrapper_open', 0 );
	// Low priority so the closing div is after any notices.
	add_action( 'admin_notices', __NAMESPACE__ . '\\add_notices_wrapper_close', 999999 );
}

/**
 * Adds an opening div to wrap around notices in the Accelerate Dashboard.
 *
 * @return void
 */
function add_notices_wrapper_open() : void {
	echo '<div id="Altis_Dashboard__notices">';

	$screen = get_current_screen();
	if ( ! Accelerate\get_config() && ! $screen->id === 'accelerate_page_altis-accelerate-settings' ) {
		get_activation_message();
	}
}

/**
 * Adds a closing div to wrap around notices in the Accelerate Dashboard.
 *
 * @return void
 */
function add_notices_wrapper_close() {
	echo '</div>';
}

/**
 * Display Accelerate React apps.
 *
 * @return void
 */
function render_page() : void {
	require_once ABSPATH . 'wp-admin/admin-header.php';

	echo '<div id="altis-analytics-root">';

	if ( wp_get_environment_type() === 'local' ) {
		echo "<p>Ensure you're running the Webpack server. You may also need to open the script URL directly to accept the SSL certificate.</p>";
	}

	echo '</div>';
	require_once ABSPATH . 'wp-admin/admin-footer.php';
}

/**
 * Register Accelerate menu pages.
 *
 * @return void
 */
function register_admin_menu() : void {
	// Scale down the admin menu image.
	add_action( 'admin_print_header_scripts', function() {
		printf( '<style>.toplevel_page_accelerate .wp-menu-image img { width: 20px; height: 20px; }</style>' );
	} );

	add_menu_page(
		__( 'Altis Accelerate', 'altis' ),
		__( 'Accelerate', 'altis' ),
		'edit_posts',
		'accelerate',
		Accelerate\get_config()
			? 'Altis\\Accelerate\\Dashboard\\load_dashboard'
			: __NAMESPACE__ . '\\render_settings_page',
		plugins_url( '../assets/altis-icon.svg', __DIR__ ),
		3
	);

	if ( ! Accelerate\get_config() ) {
		return;
	}

	add_submenu_page(
		'accelerate',
		__( 'Content Explorer', 'altis' ),
		__( 'Content Explorer', 'altis' ),
		'edit_posts',
		'accelerate',
		'Altis\\Accelerate\\Dashboard\\load_dashboard',
		0
	);
}
