<?php
/**
 * Broadcast Block Server Side.
 *
 * @phpcs:disable HM.Files.NamespaceDirectoryName.NameMismatch
 * @phpcs:disable HM.Files.FunctionFileName.WrongFile
 *
 * @package aws-analytics
 */

namespace Altis\Accelerate\Blocks\Broadcast;

use Altis\Accelerate;
use Altis\Accelerate\Blocks;
use Altis\Accelerate\Utils;

const BLOCK = 'broadcast';

/**
 * Registers the block type assets and server side rendering.
 */
function setup() {
	$block_data = Blocks\get_block_settings( BLOCK );

	// Queue up JS files.
	add_action( 'enqueue_block_editor_assets', __NAMESPACE__ . '\\enqueue_assets' );

	// Register the block.
	register_block_type( $block_data['name'], [
		'attributes' => $block_data['settings']['attributes'],
		'render_callback' => __NAMESPACE__ . '\\render_block',
	] );
}

/**
 * Enqueues the block assets.
 */
function enqueue_assets() {
	$block = Utils\register_assets(
		'blocks/broadcast',
		[
			'dependencies' => [
				'wp-api-fetch',
				'wp-plugins',
				'wp-blocks',
				'wp-i18n',
				'wp-editor',
				'wp-block-editor',
				'wp-components',
				'wp-edit-post',
				'wp-html-entities',
			],
		],
		true
	);

	wp_add_inline_script(
		$block['script'],
		sprintf(
			'window.Altis = window.Altis || {};' .
			'window.Altis.Analytics = window.Altis.Analytics || {};' .
			'window.Altis.Analytics.Broadcast = window.Altis.Analytics.Broadcast || {};' .
			'window.Altis.Analytics.Broadcast.ManagerURL = "%s";',
			'edit.php?post_type=broadcast'
		),
		'before'
	);
}

/**
 * Render callback for the Broadcast block.
 *
 * Because this block only saves <InnerBlocks.Content> on the JS side,
 * the content string represents only the wrapped inner block markup.
 *
 * @param array $attributes The block's attributes object.
 * @param string $inner_content The block's saved content.
 * @return string The final rendered block markup, as an HTML string.
 */
function render_block( array $attributes, ?string $inner_content = '' ) : string {
	$client_id = $attributes['clientId'] ?? null;
	$broadcast_id = $attributes['broadcast'] ?? null;
	$class_name = $attributes['className'] ?? '';
	$align = $attributes['align'] ?? 'none';

	// Add alignment class.
	if ( ! empty( $align ) ) {
		$class_name .= sprintf( 'align%s', $align );
	}

	$block_ids = get_post_meta( $broadcast_id, 'blocks' ) ?: [];
	$block_ids = array_map( 'absint', array_values( array_filter( $block_ids ) ) );
	$nested_blocks = array_map( function( int $block_id ) use ( $client_id ) : string {
		$block = get_post( $block_id );
		$content = $block->post_type === 'wp_block'
			? sprintf( '<!-- wp:block {"ref":%1$d} /-->', $block_id )
			: $block->post_content;
		return sprintf(
			'<template data-id="%1$d" data-parent-id="%2$s">%3$s</template>',
			$block_id,
			$client_id,
			$content
		);
	}, $block_ids );
	$inner_content = do_blocks( implode( '', $nested_blocks ) );

	// Preview.
	if ( current_user_can( 'edit_posts' ) && ( is_preview() || is_customize_preview() ) ) {
		Utils\register_assets(
			'blocks/altis-xb-preview',
			[
				'dependencies' => [
					'wp-i18n',
				],
			],
			true
		);

		return sprintf(
			'%s
			<div class="altis-xb-preview altis-xb-preview--broadcast %s" data-client-id="%s" data-post-id="%s">
				<div class="altis-xb-preview__tabs"></div>
				<div class="altis-xb-preview__content"></div>
			</div>',
			$inner_content,
			$class_name,
			$client_id,
			$broadcast_id
		);
	}

	return sprintf(
		'%4$s<broadcast-block class="%1$s" client-id="%2$s" broadcast-id="%3$d"></broadcast-block>',
		$class_name,
		$client_id,
		$broadcast_id,
		$inner_content
	);
}
