<?php
/**
 * Altis Accelerate Elasticsearch handling.
 *
 * phpcs:disable WordPress.WP.AlternativeFunctions.parse_url_parse_url
 */

namespace Altis\Accelerate\ElasticSearch;

use Altis\Accelerate;

/**
 * Set up Elasticsearch connection.
 *
 * @return void
 */
function bootstrap() : void {
	add_filter( 'altis.analytics.elasticsearch.url', function ( $url ) {
		if ( getenv( 'ACCELERATE_LOCAL_ELASTICSEARCH_URL' ) ) {
			return getenv( 'ACCELERATE_LOCAL_ELASTICSEARCH_URL' );
		}

		return $url;
	} );

	add_filter( 'altis.analytics.elasticsearch.index-prefix', function ( $index_prefix ) {
		if ( getenv( 'ACCELERATE_LOCAL_ELASTICSEARCH_URL' ) ) {
			return 'analytics-';
		}
		return $index_prefix;
	} );

	// No need for request signing.
	if ( getenv( 'ACCELERATE_LOCAL_ELASTICSEARCH_URL' ) ) {
		return;
	}

	// phpcs:ignore WordPressVIPMinimum.Hooks.RestrictedHooks.http_request_args
	add_filter( 'http_request_args', __NAMESPACE__ . '\\on_http_request_args', 10, 2 );
}

/**
 * Process HTTP request arguments.
 *
 * @param array $args Request arguments.
 * @param string $url Request URL.
 * @return array
 */
function on_http_request_args( array $args, string $url ) : array {
	$host = parse_url( $url, PHP_URL_HOST );
	$es_host = parse_url( Accelerate\get_config()['elasticsearch']['url'], PHP_URL_HOST );

	if ( $es_host !== $host ) {
		return $args;
	}

	return sign_wp_request( $args, $url );
}

/**
 * Sign requests made to Elasticsearch.
 *
 * @param array $args Request arguments.
 * @param string $url Request URL.
 * @return array
 */
function sign_wp_request( array $args, string $url ) : array {
	$config = Accelerate\get_config()['elasticsearch'];

	if ( is_array( $args['body'] ) ) {
		$args['body'] = http_build_query( $args['body'], '', '&' );
	}

	if ( isset( $args['headers']['Host'] ) ) {
		unset( $args['headers']['Host'] );
	}

	// Calculate Canonical query string per https://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html.
	parse_str( parse_url( $url, PHP_URL_QUERY ), $query );
	$canonical_query_string = '';
	ksort( $query );
	foreach ( $query as $k => $v ) {
		if ( ! is_array( $v ) ) {
			$canonical_query_string .= rawurlencode( $k ) . '=' . rawurlencode( $v !== null ? $v : '' ) . '&';
		} else {
			sort( $v );
			foreach ( $v as $value ) {
				$canonical_query_string .= rawurlencode( $k ) . '=' . rawurlencode( $value !== null ? $value : '' ) . '&';
			}
		}
	}
	$canonical_query_string = substr( $canonical_query_string, 0, -1 );

	// Calculate canonical headers.
	$exclude_headers = [
		'cache-control'         => true,
		'content-type'          => true,
		'content-length'        => true,
		'expect'                => true,
		'max-forwards'          => true,
		'pragma'                => true,
		'range'                 => true,
		'te'                    => true,
		'if-match'              => true,
		'if-none-match'         => true,
		'if-modified-since'     => true,
		'if-unmodified-since'   => true,
		'if-range'              => true,
		'accept'                => true,
		'authorization'         => true,
		'proxy-authorization'   => true,
		'from'                  => true,
		'referer'               => true,
		'user-agent'            => true,
		'X-Amz-User-Agent'      => true,
		'x-amzn-trace-id'       => true,
		'aws-sdk-invocation-id' => true,
		'aws-sdk-retry'         => true,
	];
	$canonical_headers_string = '';
	$signed_headers = '';
	$headers = $args['headers'];
	$headers['X-Amz-Date'] = gmdate( 'Ymd\THis\Z' );
	$headers['Host'] = parse_url( $url, PHP_URL_HOST );
	ksort( $headers );
	foreach ( $headers as $k => $v ) {
		$k = strtolower( $k );
		if ( ! empty( $exclude_headers[ $k ] ) ) {
			continue;
		}
		$canonical_headers_string .= $k . ':' . trim( $v ) . "\n";
		$signed_headers .= ';' . $k;
	}

	$signed_headers = substr( $signed_headers, 1 );

	$canonical_request = [
		$args['method'], // Method.
		'/' . str_replace( '%2F', '/', rawurlencode( ltrim( parse_url( $url, PHP_URL_PATH ), '/' ) ) ), // URI.
		$canonical_query_string,
		$canonical_headers_string,
		$signed_headers,
		strtolower( hash( 'sha256', $args['body'] ) ), // Hash of body.
	];

	$canonical_request = implode( "\n", $canonical_request );

	$canonical_request_hash = hash( 'sha256', utf8_encode( $canonical_request ) );

	$scope_string = substr( $headers['X-Amz-Date'], 0, 8 ) . '/' . $config['region'] . '/es/aws4_request';
	// Create v4 request string per https://docs.aws.amazon.com/general/latest/gr/sigv4-create-string-to-sign.html.
	$sign_string = [
		'AWS4-HMAC-SHA256',
		$headers['X-Amz-Date'],
		$scope_string,
		$canonical_request_hash,
	];
	$sign_string = implode( "\n", $sign_string );

	// Create signing key per https://docs.aws.amazon.com/general/latest/gr/sigv4-calculate-signature.html.
	$key_date = hash_hmac( 'sha256', substr( $headers['X-Amz-Date'], 0, 8 ), 'AWS4' . $config['access_secret'], true );
	$key_region = hash_hmac( 'sha256', $config['region'], $key_date, true );
	$key_service = hash_hmac( 'sha256', 'es', $key_region, true );
	$key_signing = hash_hmac( 'sha256', 'aws4_request', $key_service, true );
	$signature = hash_hmac( 'sha256', $sign_string, $key_signing );

	$args['headers']['Authorization'] = 'AWS4-HMAC-SHA256 Credential=' . $config['access_key'] . '/' . $scope_string . ', SignedHeaders=' . $signed_headers . ', Signature=' . $signature;
	$args['headers']['X-Amz-Date'] = $headers['X-Amz-Date'];

	return $args;
}
