<?php
/**
 * Bootstrap Altis Accelerate.
 */

namespace Altis\Accelerate;

use Altis\GlobalBlocks;

/**
 * Bootstrap function to setup hooks / load plugins.
 *
 * @return void
 */
function bootstrap() : void {
	Admin\bootstrap();

	$config = get_config();

	if ( ! $config ) {
		return;
	}

	// Low level constants for Altis core.
	if ( ! defined( 'Altis\\ROOT_DIR' ) ) {
		define( 'Altis\\ROOT_DIR', PLUGIN_DIR );
		define( 'HM_ENV', $config['elasticsearch']['index_path'] ?? get_site_url() );
		define( 'HM_ENV_ARCHITECTURE', 'accelerate' );
		define( 'HM_ENV_REGION', 'eu-central-1' );
		define( 'HM_ENV_TYPE', 'accelerate' );
		if ( ! defined( 'ALTIS_ENVIRONMENT_TIER' ) ) {
			define( 'ALTIS_ENVIRONMENT_TIER', 'free' );
		}
		if ( ! defined( 'ALTIS_FEATURE_TIER' ) ) {
			define( 'ALTIS_FEATURE_TIER', 'free' );
		}
		if ( ! defined( 'ALTIS_SUPPORT_TIER' ) ) {
			define( 'ALTIS_SUPPORT_TIER', 'default' );
		}
	}

	// Analytics service constants.
	if ( ! defined( 'ALTIS_ACCELERATE_PINPOINT_ID' ) ) {
		define( 'ALTIS_ACCELERATE_PINPOINT_ID', $config['pinpoint']['app_id'] );
		define( 'ALTIS_ACCELERATE_PINPOINT_REGION', $config['pinpoint']['region'] );
	}

	// Check for env defined pinpoint endpoint.
	if ( ! defined( 'ALTIS_ACCELERATE_PINPOINT_ENDPOINT' ) && getenv( 'ALTIS_ACCELERATE_PINPOINT_ENDPOINT' ) ) {
		define( 'ALTIS_ACCELERATE_PINPOINT_ENDPOINT', getenv( 'ALTIS_ACCELERATE_PINPOINT_ENDPOINT' ) );
	}
	if ( ! defined( 'ALTIS_ACCELERATE_PINPOINT_ENDPOINT' ) && defined( 'ALTIS_ANALYTICS_PINPOINT_ENDPOINT' ) ) {
		define( 'ALTIS_ACCELERATE_PINPOINT_ENDPOINT', ALTIS_ANALYTICS_PINPOINT_ENDPOINT );
	}


	// Configure features.
	configure_features();

	// Load libs.
	if ( ! function_exists( 'register_extended_post_type' ) ) {
		require_once __DIR__ . '/../lib/johnbillion/extended-cpts/extended-cpts.php';
	}
	require_once __DIR__ . '/../plugins/asset-loader/asset-loader.php';

	// Load telemetry features, if not provided by Altis Cloud.
	if ( ! function_exists( '\\Altis\\Telemetry\\bootstrap' ) ) {
		require_once __DIR__ . '/telemetry/namespace.php';
		Telemetry\bootstrap();
	}

	// Conditionally load demo tools.
	if ( file_exists( __DIR__ . '/../plugins/analytics-demo-tools/plugin.php' ) ) {
		require_once __DIR__ . '/../plugins/analytics-demo-tools/plugin.php';
	}

	// Rely on Altis Cloud Elasticsearch request signing and URL.
	if ( ! defined( 'ALTIS_ACCELERATE_USE_CLOUD' ) || ! ALTIS_ACCELERATE_USE_CLOUD ) {
		ElasticSearch\bootstrap();
	}

	Beta\bootstrap();
	Analytics\setup();
	API\setup();
	GlobalBlocks\bootstrap();

	if ( Utils\is_feature_enabled( 'dashboard' ) ) {
		Dashboard\setup();
	}
	if ( Utils\is_feature_enabled( 'audiences' ) ) {
		Audiences\setup();
		Preview\setup();
	}
	if ( Utils\is_feature_enabled( 'experiments' ) ) {
		Experiments\setup();
	}
	if ( Utils\is_feature_enabled( 'blocks' ) ) {
		Blocks\setup();
	}
	if ( Utils\is_feature_enabled( 'broadcast' ) ) {
		Broadcast\setup();
	}
	if ( Utils\is_feature_enabled( 'export' ) ) {
		Export\setup();
	}

	add_filter( 'altis.telemetry.env_traits', __NAMESPACE__ . '\\extend_environment_traits' );

	add_action( 'altis.analytics.enqueue_scripts', __NAMESPACE__ . '\\override_analytics_js_url' );
}

/**
 * Get the config for the plugin.
 *
 * @return array|null
 */
function get_config() : ?array {
	static $config;
	if ( $config ) {
		return $config;
	}

	if ( defined( 'ALTIS_ACCELERATE_USE_CLOUD' ) && ALTIS_ACCELERATE_USE_CLOUD ) {
		$config = [
			'elasticsearch' => [
				'url' => sprintf(
					'%s://%s%s',
					ELASTICSEARCH_PORT === 443 ? 'https' : 'http',
					ELASTICSEARCH_HOST,
					! in_array( ELASTICSEARCH_PORT, [ 80, 443 ], true ) ? ( ':' . ELASTICSEARCH_PORT ) : ''
				),
				'index_path' => 'analytics-'
			],
			'pinpoint' => [
				'app_id' => ALTIS_ANALYTICS_PINPOINT_ID,
				'region' => ALTIS_ANALYTICS_PINPOINT_REGION,
			],
		];
		return $config;
	}

	$config = get_option( 'altis_config' );
	$config = json_decode( $config, true );
	return $config;
}

/**
 * Configure default enabled features.
 *
 * @return void
 */
function configure_features() {
	add_filter( 'altis.analytics.feature.dashboard', '__return_true' );
	add_filter( 'altis.analytics.feature.audiences', '__return_true' );
	add_filter( 'altis.analytics.feature.experiments', '__return_true' );
	add_filter( 'altis.analytics.feature.blocks', '__return_true' );
	add_filter( 'altis.analytics.feature.insights', '__return_true' );
	add_filter( 'altis.analytics.feature.export', '__return_true' );
}

/**
 * Use client library for tracking script.
 *
 * @return void
 */
function override_analytics_js_url() : void {
	if ( wp_get_environment_type() === 'local' ) {
		return;
	}

	$scripts = wp_scripts();
	$altis_analytics = $scripts->query( 'altis-accelerate-accelerate' );
	if ( ! $altis_analytics ) {
		return;
	}

	$altis_analytics->src = sprintf( 'https://eu.accelerate.altis.cloud/accelerate.%s.js', VERSION );
}

/**
 * Update the environment traits for the Accelerate plugin.
 *
 * @param array $traits List of environment specific traits.
 * @return array
 */
function extend_environment_traits( array $traits ) : array {
	$traits['plugin_version'] = VERSION;
	return $traits;
}
