/**
 * Altis Analytics data store.
 */

const { apiFetch } = wp;
const { registerStore } = wp.data;
const { addQueryArgs } = wp.url;

const initialState = {
	posts: [],
	views: {},
	isLoading: false,
};

// Add API Fetch middleware to rewrite permissions lookups.
apiFetch.use( ( options, next ) => {
	if ( options.path ) {
		options.path = options.path.replace( 'wp/v2/xbs', 'accelerate/v1/xbs' );
	}
	return next( options );
} );

/**
 * Experience block redux store reducer.
 *
 * @param {Object} state Store state.
 * @param {Object} action Action object.
 * @returns {Object} Updated state.
 */
const reducer = function reducer( state, action ) {
	switch ( action.type ) {
		case 'ADD_VIEWS': {
			const key = `${ action.clientId }:${ btoa( JSON.stringify( action.args ) ) }`;
			return {
				...state,
				views: {
					...state.views,
					[ key ]: action.views,
				},
			};
		}

		case 'REMOVE_VIEWS': {
			const key = `${ action.clientId }:${ btoa( JSON.stringify( action.args ) ) }`;
			const { [ key ]: deletedItem, ...newState } = state;
			return newState;
		}

		case 'SET_IS_LOADING': {
			return {
				...state,
				isLoading: action.isLoading,
			};
		}

		case 'ADD_POST': {
			return {
				...state,
				posts: [ ...state.posts, action.post ],
			};
		}

		default: {
			return state;
		}
	}
};

const controls = {
	/**
	 * Fetch data from the API.
	 *
	 * @param {Object} action Action object.
	 * @returns {Promise} API request promise.
	 */
	FETCH_FROM_API( action ) {
		return apiFetch( action.options );
	},
	/**
	 * Convert API response to JSON object.
	 *
	 * @param {Object} action The action object.
	 * @param {Response} action.response Response object.
	 * @returns {Object} JSON data from request.
	 */
	RESPONSE_TO_JSON( action ) {
		return action.response.json();
	},
};

const actions = {
	/**
	 * Action creator for adding block analytics data.
	 *
	 * @param {string} clientId Block client ID.
	 * @param {Object} args Optional filters for the query.
	 * @param {Object} views Analytics data.
	 * @returns {Object} Redux action object.
	 */
	addViews( clientId, args, views ) {
		return {
			type: 'ADD_VIEWS',
			clientId,
			args,
			views,
		};
	},
	/**
	 * Action creator to remove block analytics data.
	 *
	 * @param {string} clientId The block client ID.
	 * @param {Object} args Optional filters for the query.
	 * @returns {Object} Redux action object.
	 */
	removeViews( clientId, args ) {
		return {
			type: 'REMOVE_VIEWS',
			clientId,
			args,
		};
	},
	/**
	 * Action creator for setting loading property.
	 *
	 * @param {boolean} isLoading True if currently loading data.
	 * @returns {Object} Redux action object.
	 */
	setIsLoading( isLoading ) {
		return {
			type: 'SET_IS_LOADING',
			isLoading,
		};
	},
	/**
	 * Action creator for adding a post to the store.
	 *
	 * @param {Object} post A post object.
	 * @returns {Object} Action creator object.
	 */
	addPost( post ) {
		return {
			type: 'ADD_POST',
			post,
		};
	},
	/**
	 * Action creator for fetching API data.
	 *
	 * @param {Object} options API fetch function options object.
	 * @returns {Object} Redux action object.
	 */
	fetch( options ) {
		return {
			type: 'FETCH_FROM_API',
			options,
		};
	},
	/**
	 * Action creator for converting Response object to JSON.
	 *
	 * @param {Reponse} response Response object.
	 * @returns {Object} Redux action object.
	 */
	json( response ) {
		return {
			type: 'RESPONSE_TO_JSON',
			response,
		};
	},
};

const selectors = {
	/**
	 * Get analytics data for a block.
	 *
	 * @param {Object} state Redux store state.
	 * @param {string} clientId Block client ID.
	 * @param {Object} args Optional args to scope views data to.
	 * @returns {Object | boolean} Object of analytics data or false on failure.
	 */
	getViews( state, clientId, args = null ) {
		return state.views[ `${ clientId }:${ btoa( JSON.stringify( args ) ) }` ] || false;
	},
	/**
	 * Get loading state for analytics data.
	 *
	 * @param {Object} state Redux store state.
	 * @returns {boolean} True if data is currently loading.
	 */
	getIsLoading( state ) {
		return state.isLoading;
	},
	/**
	 * Get an XB post from the store.
	 *
	 * @param {Object} state The current store state.
	 * @param {string} clientId The block's client ID.
	 * @returns {?Object} Post object or null.
	 */
	getPost( state, clientId ) {
		return state.posts.find( post => post.slug === clientId ) || null;
	},
};

const resolvers = {
	/**
	 * Fetch block analytics data from API.
	 *
	 * @param {string} clientId The block ID.
	 * @param {?Object} args Optional args for filtering the data.
	 * @returns {Object} Redux action object(s).
	 */
	*getViews( clientId, args = null ) {
		yield actions.setIsLoading( true );
		let path = `accelerate/v1/xbs/${ clientId }/views`;
		if ( args ) {
			path = addQueryArgs( path, args );
		}
		const response = yield actions.fetch( {
			path,
		} );
		yield actions.addViews( clientId, args, response );
		return actions.setIsLoading( false );
	},
	/**
	 * Get an experience block post.
	 *
	 * @param {string} clientId The block client ID.
	 * @returns {Object} Redux action objects.
	 */
	*getPost( clientId ) {
		yield actions.setIsLoading( true );
		try {
			const response = yield actions.fetch( {
				path: `accelerate/v1/xbs/${ clientId }`,
			} );
			yield actions.addPost( response );
		} catch ( error ) {
			yield actions.addPost( {
				slug: clientId,
				error,
			} );
		}
		return actions.setIsLoading( false );
	},
};

export const store = registerStore( 'accelerate/xbs', {
	actions,
	controls,
	initialState,
	reducer,
	resolvers,
	selectors,
} );
